/*
 * instantiate-disk.c - Create a disk entry on the manager device.
 *
 * Copyright (c) 2002-2006 Oracle Corporation.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License, version 2 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE     /* For open64() */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/statfs.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <limits.h>
#include <errno.h>

#include <linux/types.h>

#ifndef ORACLEASM_OLD_HEADERS
# include "linux/oracleasm/compat32.h"
# include "linux/oracleasm/abi.h"
# include "linux/oracleasm/disk.h"
# include "linux/oracleasm/manager.h"
#else
# include "linux/asmcompat32.h"
# include "linux/asmabi.h"
# include "linux/asmdisk.h"
# include "linux/asmmanager.h"
#endif

#include "usage.h"
#include "manager.h"

#ifndef O_DIRECT
# define O_DIRECT 00040000
#endif

enum DiskMode {
    NO_MODE = 0,
    INSTANTIATE_DISK = 1 << 0, /* oracleasm-instantiate-disk */
    CLEAN_DISK = 1 << 1,       /* oracleasm-clean-disk */
};



/*
 * Globals
 */
#define PROGNAME_INSTANTIATE "oracleasm-instantiate-disk"
#define PROGNAME_CLEAN "oracleasm-clean-disk"
const char *program_name;
const char *program_instantiate = PROGNAME_INSTANTIATE;
const char *program_clean = PROGNAME_CLEAN;

const char *program_usage;
const char *usage_instantiate = (
    "Usage: " PROGNAME_INSTANTIATE " -l <manager> <device> <label>\n"
    "       " PROGNAME_INSTANTIATE " -h\n"
    "       " PROGNAME_INSTANTIATE " -V\n"
);
const char *usage_clean = (
    "Usage: " PROGNAME_CLEAN " -l <manager> <label>\n"
    "       " PROGNAME_CLEAN " -h\n"
    "       " PROGNAME_CLEAN " -V\n"
);




/*
 * Functions
 */


#if !defined __USE_GNU && !defined __USE_XOPEN2K8
/*
 * strnlen() is useful, and it's under __USE_GNU
 */
static inline size_t strnlen(const char *s, size_t size)
{
    char *p = memchr(s, '\0', size);
    return p ? p - s : size;
}  /* strnlen() */
#endif


static int open_disk(const char *disk_name)
{
    int fd, rc;
    struct stat64 stat_buf;

    rc = stat64(disk_name, &stat_buf);
    if (rc)
        return -errno;

    /* First check, try to avoid opening tapes */
    if (!S_ISBLK(stat_buf.st_mode))
        return -ENOTBLK;

    fd = open64(disk_name, O_RDWR);
    if (fd < 0)
        return -errno;

    /* Second check, try not to write the wrong place */
    rc = fstat64(fd, &stat_buf);
    if (rc)
    {
        close(fd);
        return -errno;
    }
    
    if (!S_ISBLK(stat_buf.st_mode))
    {
        close(fd);
        return -ENOTBLK;
    }

    return fd;
}  /* open_disk() */


static int unlink_disk(const char *manager, const char *disk)
{
    int rc;
    char *asm_disk;

    asm_disk = asm_disk_path(manager, disk);
    if (!asm_disk)
        return -ENOMEM;

    rc = unlink(asm_disk);
    if (rc)
        error("Unable to remove ASM disk \"%s\": %s\n", disk,
              strerror(errno));

    free(asm_disk);
    return rc;
}  /* unlink_disk() */


static int make_disk(const char *manager, const char *disk,
                     const char *target)
{
    int rc, t_fd;
    char *asm_disk;
    struct stat64 stat_buf;

    asm_disk = asm_disk_path(manager, disk);
    if (!asm_disk)
        return -ENOMEM;

    rc = open_disk(target);
    if (rc < 0)
    {
        error("Unable to open device \"%s\": %s\n", target, strerror(-rc));
        goto out;
    }
    t_fd = rc;

    /* FIXME: Check if mounted ? */

    rc = fstat64(t_fd, &stat_buf);
    if (rc)
    {
        error("Unable to query device \"%s\": %s\n", target,
              strerror(errno));
        goto out_close;
    }

    rc = mknod(asm_disk, 0600 | S_IFBLK, stat_buf.st_rdev);
    if (rc)
    {
        error("Unable to create ASM disk \"%s\": %s\n", disk,
              strerror(errno));
        goto out_close;
    }

    rc = open_disk(asm_disk);
    if (rc < 0)
    {
        error("Unable to open ASM disk \"%s\": %s\n", disk,
              strerror(-rc));
        /* Ignore the result, let the above error fall through */
        unlink_disk(manager, disk);
    }
    else
    {
        close(rc);
        rc = 0;
    }

out_close:
    close(t_fd);

out:
    free(asm_disk);
    return rc;
}  /* make_disk() */


extern char *optarg;
extern int optopt;
extern int opterr;
static int parse_options(int argc, char *argv[], int mode, char **manager,
                         char **target, char **label)
{
    int c;

    opterr = 0;
    while((c = getopt(argc, argv, ":hV-:l:")) != EOF)
    {
        switch(c)
        {
            case 'h':
                usage(0, NULL);
                break;

            case 'V':
                version();
                break;

            case '-':
                if (!strcmp(optarg, "help"))
                    usage(0, NULL);
                else if (!strcmp(optarg, "version"))
                    version();
                else
                    usage(-EINVAL, "Invalid option: \'--%s\'\n", optarg);
                break;

            case 'l':
                if (!optarg || !*optarg)
                    usage(-EINVAL, "Argument to \'-l\' cannot be \"\"\n");

                *manager = optarg;
                break;

            case '?':
                usage(-EINVAL, "Invalid option: \'-%c\'\n", optopt);
                break;

            case ':':
                usage(-EINVAL, "Option \'-%c\' requires an argument\n",
                      optopt);
                break;

            default:
                usage(-EINVAL, "%s: Can't get here!\n", __func__);
                break;
        }
    }

    if (mode & INSTANTIATE_DISK)
    {
        if (optind >= argc)
            usage(-EINVAL, "You must specify a target device\n");

        *target = argv[optind];
        optind++;
    }

    if (optind >= argc)
        usage(-EINVAL, "You must specify a label\n");

    *label = argv[optind];
    optind++;

    c = asmdisk_toupper(*label, -1, 0);
    if (c)
        usage(-EINVAL, "Invalid characters in label: \'%c\'\n", c);

    if (optind < argc)
        usage(-EINVAL, "Too many arguments\n");

    return 0;
}  /* parse_options() */




/*
 * Main program
 */
int main(int argc, char *argv[])
{
    char *cmd = argv[0];
    char *slash = strrchr(cmd, '/');
    int rc, mode = NO_MODE;
    char *manager = NULL, *label = NULL, *target = NULL;

    if (slash)
    {
        *slash++ = 0;
        cmd = slash;
    }

    program_name = cmd;
    argv[0] = cmd;

    if (!strcmp(cmd, program_instantiate))
    {
        mode = INSTANTIATE_DISK;
        program_usage = usage_instantiate;
    }
    else if (!strcmp(cmd, program_clean))
    {
        mode = CLEAN_DISK;
        program_usage = usage_clean;
    }
    else
    {
        rc = -ENOENT;
        program_name = cmd;
        error("disk.c cannot handle \"%s\" internally\n", cmd);
        goto out;
    }

    rc = parse_options(argc, argv, mode, &manager, &target, &label);
    if (rc)
        goto out;

    if (!manager || !*manager)
        usage(-EINVAL, "You must specify a manager.\n");
    rc = is_manager(manager);
    if (rc)
    {
        error("Unable to open manager: %s\n", strerror(-rc));
        goto out;
    }

    if (mode & INSTANTIATE_DISK)
        rc = make_disk(manager, label, target);
    else if (mode & CLEAN_DISK)
        rc = unlink_disk(manager, label);

out:
    return rc;
}  /* main() */


